import pytest

from migri.backends.postgresql import PostgreSQLConnection
from migri.elements import Query

q1 = Query("INSERT INTO mytable (a) VALUES ($a), ($b)", values={"a": 150, "b": 300})

q2 = Query("UPDATE tbl SET info=$info WHERE id=$id", values={"id": 39, "info": "ok"})

q3 = Query("SELECT * FROM pg_type")

q4 = Query(
    "SELECT * FROM val WHERE (value < $value_a AND status = $status) "
    "OR (value > $value_b AND status = $status)",
    values={"value_a": 20, "value_b": 100, "status": "ok"},
)


@pytest.mark.parametrize(
    "query_element,expected_query,expected_value",
    [
        (q1, "INSERT INTO mytable (a) VALUES ($1), ($2)", [150, 300]),
        (q2, "UPDATE tbl SET info=$2 WHERE id=$1", [39, "ok"]),
        (q3, "SELECT * FROM pg_type", []),
        (
            q4,
            "SELECT * FROM val WHERE (value < $1 AND status = $3) "
            "OR (value > $2 AND status = $3)",
            [20, 100, "ok"],
        ),
    ],
)
def test_compile(query_element, expected_query, expected_value):
    backend = PostgreSQLConnection("postgres")
    assert backend._compile(query_element) == {
        "query": expected_query,
        "values": expected_value,
    }



import glob
import importlib.util
import itertools
import logging
import os
from dataclasses import dataclass
from datetime import datetime, timezone
from inspect import iscoroutinefunction
from typing import AsyncGenerator, Iterable, List, Optional, Union

import sqlparse

from migri.elements import Query
from migri.interfaces import Task

__all__ = ["Initialize", "Migrate"]
logger = logging.getLogger(__name__)

MIGRATION_TABLE_NAME = "applied_migration"


@dataclass
class Migration:
    abspath: str
    name: Union[str, None] = None
    file_ext: Union[str, None] = None

    def __post_init__(self):
        name = os.path.basename(self.abspath)
        self.name, self.file_ext = os.path.splitext(name)


class MigrationFailed(Exception):
    ...


class MigrationOk(Exception):
    ...


class MigrationFilesMixin(object):
    MIGRATION_FILE_EXTENSIONS = ["py", "sql"]

    def _find_migrations(self, migrations_path: str) -> Iterable[str]:
        return itertools.chain(
            *(
                glob.iglob(f"{migrations_path}/*.{ext}")
                for ext in self.MIGRATION_FILE_EXTENSIONS
            )
        )

    def get_migrations(self, migrations_dir: str) -> List[Migration]:
        path = os.path.abspath(migrations_dir)

        if not os.path.isdir(path):
            raise NotADirectoryError(f"Migrations dir not found: {path}")

        return [Migration(abspath=p) for p in sorted(self._find_migrations(path))]


class MigrationApplyMixin(object):
    async def _apply_migration_from_module(self, path: str) -> bool:
        spec = importlib.util.spec_from_file_location("migration", path)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        migrate_func = getattr(module, "migrate", None)

        if not migrate_func:
            raise ImportError(f"Module '{path}' has no function migrate()")
        else:
            if iscoroutinefunction(migrate_func):
                return await migrate_func(self._connection.database)
            else:
                raise RuntimeError("migrate() expected to be an async function")

    async def _apply_migration_from_sql_file(self, path: str) -> bool:
        with open(path, "r") as f:
            contents = f.read()
            statements = filter(lambda s: s != "", sqlparse.split(contents))

        try:
            for statement in statements:
                await self._connection.execute(Query(statement))
        except Exception as e:
            logger.warning("Error running migration %s: %s", path, e)
            raise MigrationFailed from e

        return True


class Initialize(MigrationApplyMixin, MigrationFilesMixin, Task):
    async def run(self):
        """Create 'applied_migration' table"""
        migration_table_file_path = os.path.join(
            os.path.dirname(__file__), "sql/applied_migration.sql"
        )

        # Create table
        transaction = self._connection.transaction()

        async with transaction:
            await self._apply_migration_from_sql_file(migration_table_file_path)
            await transaction.commit()


class Migrate(MigrationApplyMixin, MigrationFilesMixin, Task):
    async def _apply_migrations(
        self, migrations: List[Migration]
    ) -> AsyncGenerator[str, None]:
        """Apply migrations and return list of migrations that were applied"""
        for migration in migrations:
            # Apply migrations
            if migration.file_ext == ".py":
                migration_status = await self._apply_migration_from_module(
                    migration.abspath
                )
            elif migration.file_ext == ".sql":
                migration_status = await self._apply_migration_from_sql_file(
                    migration.abspath
                )
            else:
                migration_status = False

            # Update migration record
            if migration_status is True:
                await self._record_migration(migration)

                yield migration.name

    def _pre_migration_checks(self, migrations: List[Migration]) -> bool:
        """Check migrations before applying"""
        if len(migrations) == 0:
            raise MigrationOk("All synced! No new migrations to apply! 🥳")

        return True

    async def _migrations_to_apply(
        self, migrations: List[Migration]
    ) -> List[Migration]:
        """Takes migration paths and uses migration file names to search for entries in
        'applied_migration' table
        """
        to_apply = []

        for migration in migrations:
            query = Query(
                f"SELECT id FROM {MIGRATION_TABLE_NAME} WHERE name = $migration_name",
                values={"migration_name": migration.name},
            )
            applied_migration = await self._connection.fetch_all(query)

            if not applied_migration:
                to_apply.append(migration)

        return to_apply

    async def _record_migration(self, migration: Migration):
        query = Query(
            f"INSERT INTO {MIGRATION_TABLE_NAME} (date_applied, name) "
            f"VALUES ($date, $migration_name)",
            values={
                "date": datetime.now(tz=timezone.utc),
                "migration_name": migration.name,
            },
        )

        await self._connection.execute(query)

    async def _apply(self, migrations: List[Migration], dry_run: bool):
        transaction = self._connection.transaction()
        await transaction.start()

        try:
            async for migration in self._apply_migrations(migrations):
                self.echo.info(f"Applied migration: {migration}")
        except Exception:
            logger.exception("Rolled back migration due to an error.")
            await transaction.rollback()
            raise
        else:
            # Roll back transaction if dry run mode
            # Otherwise, commit
            if dry_run:
                await transaction.rollback()
                self.echo.info("Successfully applied migrations in dry run mode.")
            else:
                await transaction.commit()

    async def run(
        self, migrations_dir: str, dry_run: Optional[bool] = False,
    ):
        migrations = self.get_migrations(migrations_dir)

        if not migrations:
            self.echo.info("No migrations to apply. Migrations directory is empty.")
            return

        migrations = await self._migrations_to_apply(migrations)

        # Apply migrations
        try:
            self._pre_migration_checks(migrations)
        except MigrationOk as e:
            self.echo.info(e)
        else:
            await self._apply(migrations, dry_run)



import asyncio
import logging
import os
import sys
from importlib import import_module
from typing import Optional, Type, TYPE_CHECKING, Union

import click

if TYPE_CHECKING:
    from asyncpg import Connection

from migri import migration
from migri.interfaces import ConnectionBackend
from migri.utils import deprecated, Echo

DEFAULT_LOG_LEVEL = "error"
LEGACY_FUNCTIONALITY_END_OF_LIFE = "1.1.0"
SUPPORTED_DIALECTS = {
    "mysql": "migri.backends.mysql::MySQL",
    "postgresql": "migri.backends.postgresql::PostgreSQL",
    "sqlite": "migri.backends.sqlite::SQLite",
}

logging.basicConfig(
    format="%(asctime)s\t%(levelname)s: %(message)s",
    datefmt="%Y-%m-%d %I:%M:%S%z",
    level=os.getenv("LOG_LEVEL", DEFAULT_LOG_LEVEL).upper(),
)
logger = logging.getLogger(__name__)


async def run_initialization(*args, **kwargs):
    """No longer supported but left here for compatibility"""
    message = (
        "Command `init` and run_initialization() are no longer supported and are now "
        "handled automatically by `migrate` and run_migrations(). See README."
    )
    deprecated(message, LEGACY_FUNCTIONALITY_END_OF_LIFE)
    Echo.info(message)


async def run_migrations(
    migrations_dir: str,
    conn: "Connection" = None,
    db_user: str = None,
    db_pass: str = None,
    db_name: str = None,
    db_host: str = None,
    db_port: str = None,
    dry_run: bool = False,
    force_close_conn: bool = True,
):
    message = (
        "run_migrations() is about to be deprecated. Please use "
        "apply_migrations(). See README."
    )
    deprecated(message, LEGACY_FUNCTIONALITY_END_OF_LIFE)
    Echo.info(message)

    # Rig new functionality to provide backwards compatibility
    conn_backend = _get_backend(SUPPORTED_DIALECTS["postgresql"])

    if not conn:
        conn = conn_backend(
            db_name, db_user=db_user, db_pass=db_pass, db_host=db_host, db_port=db_port,
        )
    else:
        conn = conn_backend("postgres", db=conn)

    # Now call new migration interface 💪
    await apply_migrations(migrations_dir, conn, dry_run, force_close_conn)


async def apply_migrations(
    migrations_dir: str,
    conn: ConnectionBackend,
    dry_run: bool = False,
    force_close_conn: bool = True,  # TODO For backwards compatibility, remove in 1.1.0
):
    init_task = migration.Initialize(conn)
    migrate_task = migration.Migrate(conn)

    await conn.connect()
    await init_task.run()
    await migrate_task.run(migrations_dir, dry_run)

    if force_close_conn:
        await conn.disconnect()


def _get_backend(module_info: str) -> Type[ConnectionBackend]:
    module_name, module_class_prefix = module_info.split("::")
    module = import_module(module_name)
    return getattr(module, f"{module_class_prefix}Connection")


def get_connection(
    db_name: str,
    db_user: Optional[str] = None,
    db_pass: Optional[str] = None,
    db_host: Optional[str] = None,
    db_port: Optional[Union[int, str]] = None,
    dialect: Optional[str] = None,
) -> ConnectionBackend:
    """Infer db dialect if not provided and initialize connection to database"""
    # If no dialect, infer
    if not dialect:
        if not db_port:
            dialect = "sqlite"
        elif int(db_port) == 3306:
            dialect = "mysql"
        elif int(db_port) == 5432:
            dialect = "postgresql"
        else:
            raise RuntimeError(
                "Unable to infer database dialect, please specify dialect"
            )

    try:
        module_info = SUPPORTED_DIALECTS[dialect]
    except KeyError:
        raise RuntimeError(f"The dialect '{dialect}' is not supported")

    connection = _get_backend(module_info)

    return connection(
        db_name, db_user=db_user, db_pass=db_pass, db_host=db_host, db_port=db_port,
    )


@click.group()
@click.option("-n", "--db-name", required=True, default=lambda: os.getenv("DB_NAME"))
@click.option("-u", "--db-user", default=lambda: os.getenv("DB_USER"))
@click.option("-s", "--db-pass", default=lambda: os.getenv("DB_PASS"))
@click.option("-h", "--db-host", default=lambda: os.getenv("DB_HOST"))
@click.option("-p", "--db-port", default=lambda: os.getenv("DB_PORT"))
@click.option("-d", "--dialect", default=lambda: os.getenv("DB_DIALECT"))
@click.option(
    "-l", "--log-level", default=lambda: os.getenv("LOG_LEVEL", DEFAULT_LOG_LEVEL)
)
@click.pass_context
def cli(ctx, **kwargs) -> None:
    log_level = kwargs.pop("log_level")

    logger.setLevel(log_level.upper())
    logger.debug("Log %s", logging.getLevelName(log_level))

    # Expose db creds to commands via context
    ctx.ensure_object(dict)
    ctx.obj["connection"] = get_connection(
        db_name=kwargs["db_name"],
        db_user=kwargs["db_user"],
        db_pass=kwargs["db_pass"],
        db_host=kwargs["db_host"],
        db_port=kwargs["db_port"],
        dialect=kwargs["dialect"],
    )


@cli.command(short_help="Create migrations table to begin using migri [unsupported]")
@click.pass_context
def init(ctx) -> None:
    asyncio.run(run_initialization())


@cli.command(short_help="Run all unapplied migrations in lexicographical order")
@click.option(
    "-m",
    "--migrations-dir",
    required=True,
    default=lambda: os.getenv("MIGRATIONS_DIR", "migrations"),
)
@click.option("--dry-run", default=False, is_flag=True)
@click.pass_context
def migrate(ctx, migrations_dir: str, dry_run: bool) -> None:
    asyncio.run(apply_migrations(migrations_dir, ctx.obj["connection"], dry_run))


def main():
    try:
        cli()
    except Exception:
        logger.exception(
            "Caught exception in migri.main::main that caused migri to exit"
        )
        sys.exit(1)



from typing import Any, Dict, List, Optional, TYPE_CHECKING, Union

if TYPE_CHECKING:
    from aiomysql import Connection as MySQLConnection
    from aiosqlite import Connection as SQLiteConnection
    from asyncpg import Connection as PostgreSQLConnection

from migri.elements import Query
from migri.utils import Echo

Database = Union["MySQLConnection", "PostgreSQLConnection", "SQLiteConnection"]


class ConnectionBackend:
    def __init__(
        self,
        db_name: str,
        db_user: Optional[str] = None,
        db_pass: Optional[str] = None,
        db_host: Optional[str] = None,
        db_port: Optional[Union[int, str]] = None,
        db: Optional[
            Database
        ] = None,  # For providing backwards compatibility, to be removed in 1.1.0
    ):
        self._db_user = db_user
        self._db_pass = db_pass
        self._db_name = db_name
        self._db_host = db_host
        self._db_port = db_port
        self._db = db  # TODO default to None in 1.1.0

    async def __aenter__(self) -> "ConnectionBackend":
        await self.connect()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.disconnect()
        return self

    async def connect(self):
        raise NotImplementedError

    @property
    def database(self) -> Database:
        return self._db

    async def disconnect(self):
        raise NotImplementedError

    async def execute(self, query: Query):
        raise NotImplementedError

    async def fetch(self, query: Query) -> Dict[str, Any]:
        raise NotImplementedError

    async def fetch_all(self, query: Query) -> List[Dict[str, Any]]:
        raise NotImplementedError

    def transaction(self) -> "TransactionBackend":
        raise NotImplementedError


class TransactionBackend:
    def __init__(self, connection: ConnectionBackend):
        self._connection = connection

    async def __aenter__(self) -> "TransactionBackend":
        raise NotImplementedError

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        raise NotImplementedError

    async def start(self):
        raise NotImplementedError

    async def commit(self):
        raise NotImplementedError

    async def rollback(self):
        raise NotImplementedError


class Task:
    def __init__(self, connection: ConnectionBackend):
        self.echo = Echo
        self._connection = connection

    async def run(self, *args, **kwargs):
        raise NotImplementedError



import asyncpg
import re
from typing import Any, Dict, List

from migri.elements import Query
from migri.interfaces import ConnectionBackend, TransactionBackend


class PostgreSQLConnection(ConnectionBackend):
    @staticmethod
    def _compile(query: Query) -> dict:
        q = query.statement
        v = []

        if query.placeholders:
            keys = list(query.values.keys())
            v = [query.values[k] for k in keys]

            for p in query.placeholders:
                # Find index of key and add 1
                replacement = f"${keys.index(p.replace('$', '')) + 1}"

                # Substitute
                q = re.sub(f"\\{p}", replacement, q)

        return {"query": q, "values": v}

    async def connect(self):
        if not self._db:
            self._db = await asyncpg.connect(
                host=self._db_host,
                port=self._db_port,
                user=self._db_user,
                password=self._db_pass,
                database=self._db_name,
            )

    async def disconnect(self):
        await self._db.close()

    async def execute(self, query: Query):
        q = self._compile(query)
        await self._db.execute(q["query"], *q["values"])

    async def fetch(self, query: Query) -> Dict[str, Any]:
        q = self._compile(query)
        res = await self._db.fetchrow(q["query"], *q["values"])
        return dict(res)

    async def fetch_all(self, query: Query) -> List[Dict[str, Any]]:
        q = self._compile(query)
        res = await self._db.fetch(q["query"], *q["values"])

        return [dict(r) for r in res]

    def transaction(self) -> TransactionBackend:
        return PostgreSQLTransaction(self)


class PostgreSQLTransaction(TransactionBackend):
    def __init__(self, connection: ConnectionBackend):
        super().__init__(connection)
        self._committed = False
        self._transaction = None

    async def __aenter__(self) -> TransactionBackend:
        await self.start()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.rollback()

    async def start(self):
        self._transaction = self._connection.database.transaction()
        await self._transaction.start()

    async def commit(self):
        await self._transaction.commit()
        self._committed = True

    async def rollback(self):
        if not self._committed:
            await self._transaction.rollback()



import warnings
from typing import Optional

import click


class Echo:
    @classmethod
    def error(cls, message: str):
        click.secho(message, bold=True, fg="red")

    @classmethod
    def info(cls, message: str):
        click.echo(message)

    @classmethod
    def success(cls, message: str):
        click.secho(message, bold=True)


def deprecated(message: str, end_of_life: Optional[str] = None):
    """Use to warn of deprecation or pending deprecation. If end_of_life is provided,
    will append message with version in which functionality will be deprecated.
    And PendingDeprecationWarning will be used. Otherwise, DeprecationWarning will be
    used.

    :param message: Deprecation message
    :type message: str
    :param end_of_life: Version in which functionality will be deprecated
    :type end_of_life: str, optional
    """
    if end_of_life:
        message = f"{message} Will be removed in {end_of_life}."
        warning = PendingDeprecationWarning
    else:
        warning = DeprecationWarning

    warnings.warn(message, warning, stacklevel=2)



import asyncio
from contextlib import asynccontextmanager
from typing import Generator

import asyncpg
import pytest

from migri import get_connection
from migri.elements import Query

TEST_DB_CREDS = {
    "db_user": "migrator",
    "db_pass": "passpass",
    "db_name": "migritestdb",
    "db_host": "localhost",
    "db_port": "5432",
}


@asynccontextmanager
async def db_conn(
    db_user: str, db_pass: str, db_name: str, db_host: str, db_port: str
) -> Generator[asyncpg.Connection, None, None]:
    conn = await asyncpg.connect(
        f"postgresql://{db_user}:{db_pass}@{db_host}:{db_port}/{db_name}"
    )

    try:
        yield conn
    finally:
        await conn.close()


@pytest.yield_fixture
async def conn() -> Generator[asyncpg.connection.Connection, None, None]:
    async with db_conn(**TEST_DB_CREDS) as conn:
        tr = conn.transaction()
        await tr.start()
        yield conn

        if not conn.is_closed():
            await tr.rollback()


@pytest.fixture
def get_postgresql_conn():
    def _wrapped():
        return get_connection(
            TEST_DB_CREDS["db_name"],
            db_user=TEST_DB_CREDS["db_user"],
            db_pass=TEST_DB_CREDS["db_pass"],
            db_host=TEST_DB_CREDS["db_host"],
            db_port=TEST_DB_CREDS["db_port"],
        )

    return _wrapped


@pytest.yield_fixture
async def postgresql_conn_factory(get_postgresql_conn):
    try:
        yield get_postgresql_conn
    finally:
        conn = get_postgresql_conn()

        async with conn:
            await conn.execute(Query("DROP SCHEMA public CASCADE"))
            await conn.execute(Query("CREATE SCHEMA public"))


async def execute(command: str, dbname: str) -> str:
    async with db_conn(**{**TEST_DB_CREDS, "db_name": dbname}) as conn:
        return await conn.execute(command)


def pytest_configure(config):
    asyncio.run(execute("CREATE DATABASE migritestdb", "postgres"))


def pytest_unconfigure(config):
    asyncio.run(execute("DROP DATABASE migritestdb", "postgres"))



from setuptools import setup, find_packages


with open("README.md", "r", encoding="utf-8") as f:
    readme = f.read()

setup(
    name="migri",
    version="0.3.0",
    long_description=readme,
    long_description_content_type="text/markdown",
    extras_requires={
        # "mysql": ["aiomysql"],  TODO enable when ready
        "postgresql": ["asyncpg"],
        # "sqlite": ["aiosqlite"],  TODO enable when ready
    },
    install_requires=[
        "asyncpg",  # TODO remove in 1.1.0
        "click==7.*",
        "sqlparse==0.3.*",
    ],
    entry_points={
        "console_scripts": [
            "migri = migri.main:main"
        ]
    },
    packages=find_packages(),
    include_package_data=True,
    python_requires=">=3.7",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Database",
    ]
)



import pytest
from asyncpg import InterfaceError
from datetime import datetime
from freezegun import freeze_time

from migri import apply_migrations, run_migrations
from migri.elements import Query

MIGRATIONS_A_DIR = "test/migrations_a"
MIGRATIONS_B_DIR = "test/migrations_b"
MIGRATIONS_C_DIR = "test/migrations_c"

pytestmark = pytest.mark.asyncio


@freeze_time("2020-7-3 22:00:01")
async def test_apply_migrations_successful(capsys, postgresql_conn_factory):
    """Apply migrations and then check that friendly message is output if
    migrations are up-to-date"""
    # Apply migrations
    conn = postgresql_conn_factory()
    await apply_migrations(MIGRATIONS_A_DIR, conn)

    # Check that record table was created
    record_query = Query(
        "SELECT column_name FROM information_schema.columns "
        "WHERE table_schema = $schema AND table_name = $table_name",
        values={"schema": "public", "table_name": "record"},
    )

    conn = postgresql_conn_factory()

    async with conn:
        record_columns = await conn.fetch_all(record_query)

    assert len(record_columns) == 2
    assert record_columns[0]["column_name"] == "id"
    assert record_columns[1]["column_name"] == "user_id"

    # Check that migrations were recorded
    conn = postgresql_conn_factory()

    async with conn:
        migrations_query = Query("SELECT * FROM applied_migration")
        applied_migrations = await conn.fetch_all(migrations_query)

    assert len(applied_migrations) == 3

    for migration in applied_migrations:
        assert (
            migration["date_applied"].isoformat()
            == f"{datetime.utcnow().isoformat()}+00:00"
        )

    # Check output
    captured = capsys.readouterr()
    expected_output = (
        "Applied migration: 0001_initial\n"
        "Applied migration: 0002_add_initial_data\n"
        "Applied migration: 0003_record\n"
    )

    assert captured.out == expected_output
    assert captured.err == ""

    # Apply migrations again - we expect friendly output to let us know that migrations
    # are all up to date
    conn = postgresql_conn_factory()

    await apply_migrations(MIGRATIONS_A_DIR, conn)

    captured = capsys.readouterr()
    expected_output = "All synced! No new migrations to apply! 🥳\n"

    assert captured.out == expected_output
    assert captured.err == ""


async def test_apply_migrations_none(capsys, postgresql_conn_factory):
    """Ensure friendly message is output when migrations directory is empty"""
    # Apply migrations
    conn = postgresql_conn_factory()
    await apply_migrations(MIGRATIONS_C_DIR, conn)

    # Check that there aren't any tables
    conn = postgresql_conn_factory()

    async with conn:
        tables_query = Query(
            "SELECT table_name FROM information_schema.tables"
            " WHERE table_schema='public' AND table_type='BASE TABLE';"
        )
        tables = await conn.fetch_all(tables_query)

    assert len(tables) == 1  # applied_migration expected
    assert tables[0]["table_name"] == "applied_migration"

    # Check output
    captured = capsys.readouterr()
    expected_output = "No migrations to apply. Migrations directory is empty.\n"

    assert captured.out == expected_output
    assert captured.err == ""


async def test_apply_migrations_dry_run(capsys, postgresql_conn_factory):
    # Apply migrations in dry run mode
    conn = postgresql_conn_factory()
    await apply_migrations(MIGRATIONS_A_DIR, conn, dry_run=True)

    # Check that there aren't any tables
    conn = postgresql_conn_factory()

    async with conn:
        tables_query = Query(
            "SELECT table_name FROM information_schema.tables"
            " WHERE table_schema='public' AND table_type='BASE TABLE';"
        )
        tables = await conn.fetch_all(tables_query)

    assert len(tables) == 1  # applied_migration expected
    assert tables[0]["table_name"] == "applied_migration"

    # Check output
    captured = capsys.readouterr()
    expected_output = (
        "Applied migration: 0001_initial\n"
        "Applied migration: 0002_add_initial_data\n"
        "Applied migration: 0003_record\n"
        "Successfully applied migrations in dry run mode.\n"
    )

    assert captured.out == expected_output
    assert captured.err == ""


async def test_apply_migrations_with_empty_statement_successful(
    capsys, postgresql_conn_factory
):
    """
    If sql file ends w/ empty line, sqlparse returns an empty string as a statement.
    Test that migration goes through and empty statement is filtered out
    """
    conn = postgresql_conn_factory()
    await apply_migrations(MIGRATIONS_B_DIR, conn)

    # Query for tables
    conn = postgresql_conn_factory()

    async with conn:
        tables_query = Query(
            "SELECT table_name FROM information_schema.tables"
            " WHERE table_schema='public' AND table_type='BASE TABLE';"
        )
        tables = await conn.fetch_all(tables_query)

    assert len(tables) == 3

    for table in tables:
        assert table["table_name"] in [
            "applied_migration",
            "state_machine",
            "state_history",
        ]

    # Check output
    captured = capsys.readouterr()
    expected_output = "Applied migration: 0001_initial\n"

    assert captured.out == expected_output
    assert captured.err == ""


# TODO remove in 1.1.0
async def test_migrations_no_close_conn(postgresql_conn_factory):
    """force_close_conn is a feature for backwards compatibility. Defaults to true"""
    conn = postgresql_conn_factory()
    await apply_migrations(MIGRATIONS_A_DIR, conn, force_close_conn=False)

    # Check that underlying connection can still be used
    result = await conn.database.fetchrow(
        "SELECT * FROM account WHERE name = $1", "My Account"
    )

    assert result

    await conn.disconnect()


@freeze_time("2019-10-7 19:00:01")
async def test_deprecated_run_migrations_successful(postgresql_conn_factory, conn):
    await run_migrations(MIGRATIONS_A_DIR, conn, force_close_conn=False)

    # Check account
    account = await conn.fetchrow("SELECT * FROM account WHERE name = $1", "My Account")

    assert account["name"] == "My Account"

    # Check user
    user = await conn.fetchrow("SELECT * FROM app_user")

    assert user["account_id"] == account["id"]
    assert user["first_name"] == "M"
    assert user["last_name"] == "Test"
    assert user["email"] == "test@example.com"
    assert user["password"] == "078bbb4bf0f7117fb131ec45f15b5b87"

    # Check that record table was created
    record_columns = await conn.fetch(
        "SELECT column_name FROM information_schema.columns "
        "WHERE table_schema = $1 AND table_name = $2",
        "public",
        "record",
    )

    assert len(record_columns) == 2
    assert record_columns[0]["column_name"] == "id"
    assert record_columns[1]["column_name"] == "user_id"

    # Check that migrations were recorded
    applied_migrations = await conn.fetch("SELECT * FROM applied_migration")

    assert len(applied_migrations) == 3

    for migration in applied_migrations:
        assert (
            migration["date_applied"].isoformat()
            == f"{datetime.utcnow().isoformat()}+00:00"
        )


async def test_deprecated_run_migrations_dry_run(postgresql_conn_factory, conn):
    """If running in dry run mode, pending migration(s) won't be committed. Useful for
    testing that migrations are error-free"""
    await run_migrations(MIGRATIONS_A_DIR, conn, dry_run=True, force_close_conn=False)

    tables = await conn.fetch(
        "SELECT table_name FROM information_schema.tables"
        " WHERE table_schema='public' AND table_type='BASE TABLE';"
    )

    assert len(tables) == 1  # applied_migration expected
    assert tables[0]["table_name"] == "applied_migration"


async def test_deprecated_run_migrations_with_empty_statement_successful(
    postgresql_conn_factory, conn
):
    """
    If sql file ends w/ empty line, sqlparse returns an empty string as a statement.
    Test that migration goes through and empty statement is filtered out
    """
    await run_migrations(MIGRATIONS_B_DIR, conn, force_close_conn=False)

    tables = await conn.fetch(
        "SELECT table_name FROM information_schema.tables WHERE table_schema='public'"
    )

    assert len(tables) == 3

    for table in tables:
        assert table["table_name"] in [
            "applied_migration",
            "state_machine",
            "state_history",
        ]


async def test_deprecated_run_migrations_close_conn(postgresql_conn_factory, conn):
    """By default, passed in conn should be closed"""
    await run_migrations(MIGRATIONS_A_DIR, conn)

    with pytest.raises(InterfaceError):
        await conn.fetchrow("SELECT * FROM account WHERE name = $1", "My Account")



import re
from typing import Any, Dict, Optional, Set

Values = Optional[Dict[str, Any]]


class Query:
    """Query element. Use $-based substitutions for safe parameter substitutions across
    databases.

    :param statement: Query statement in `string.Template` format
        (e.g. "SELECT * FROM applied_migration WHERE id = $id)
    :type statement: str
    :param values: Dictionary of values that will be used to substitute placeholders
        (e.g. {"id": 2})
    :type values: dict, optional
    """

    def __init__(self, statement: str, values: Values = None):
        self._statement = statement
        self._values = values

    @property
    def placeholders(self) -> Set[str]:
        matches = re.findall(r"\$(?:[_a-z][_a-z0-9]*)", self._statement)

        return set(matches)

    @property
    def statement(self) -> str:
        return self._statement

    @property
    def values(self) -> Values:
        return self._values
